# PIG Pilot
# Copyright (C) 2005  Robin J Patenall
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#

# Basic system imports
import sys

# Pygame imports
import pygame

# PyOpenGL imports
try:
    from OpenGL.GL import *
    from OpenGL.GLU import *
except:
    print 'This entry requires PyOpenGL'
    raise SystemExit

# Import the update handlers
import update

# Let's initiate pygame and SDL.
pygame.init()

# Initially force the game size to something, possibly make this a command
# line argument.
screen_size = screen_width, screen_height = 640, 480

# Initially force the game into a window, possibly make this a command line
# argument.
screen_windowed = True

# Initially set the smallest frame time to 20ms (50fps) possibly make this a
# command line argument.
time_min_frame = 20

# Try to set up the screen.
screen_flags = pygame.OPENGL | pygame.DOUBLEBUF;
if not screen_windowed :
  screen_flags |= pygame.FULLSCREEN
screen = pygame.display.set_mode(screen_size, screen_flags)

# Set a caption. Replace this with a better one!!!
pygame.display.set_caption("PIG Pilot")

# The current update handler.
update_handler = update.Menu() #update.Wait_For_ESC()

# Main event loop. Get an old time for starting off.
time_old = pygame.time.get_ticks()
events = []
while 1 :
  
  # If there isn't an update handler ...
  if update_handler == None :
    
    # ... set a default frame time.
    time_min_frame = 20
    
    # Ensure that we get any quit events
    events += pygame.event.get()
  else :
            
    # If the update handler has a frame delay ...
    if hasattr(update_handler, 'frame_delay') :
    
      # ... store it.
      time_min_frame = update_handler.frame_delay
    else :
      
      # No delay. set a default.
      time_min_frame = 20
    
    # If we are event based ...
    if time_min_frame < 0 :
      
      # ... clear this for later.
      time_min_frame = 20
      
      # ... wait for an event
      events += [pygame.event.wait()]
               
    # Get any other events
    events += pygame.event.get()
  
  # If there is a quit event then stop.
  for event in events :
    if(event.type == pygame.QUIT) :
      sys.exit()
  
  # Get the current frame time.
  time_now = pygame.time.get_ticks()
  
  # Compute the frame time.
  time_frame = time_now - time_old
      
  # To reduce the cpu load limit, if the frame time is less than the 
  # minimum ...
  if(time_frame < time_min_frame) :
    
    # ... wait for a bit ...
    pygame.time.wait(time_min_frame - time_frame);     
    
    # ... and try the main loop again.
    continue
  
  # Reset the old time for the noxt frame.
  time_old = time_now
  
  # Convert the frame time into seconds
  time_frame *= 0.001
  
  # If there isn't an update handler ...
  if update_handler == None :
  
    # ... just clear the screen.
    glClearColor(0.0, 0.0, 1.0, 0.0)
    glClear(GL_COLOR_BUFFER_BIT)

  else :
    
    # Run the handler.
    handler_new = update_handler.update(screen_size, time_frame, 
                                        events)
                                               
    # If there is a new update handler ...
    if handler_new != None :
      
      # ... delete the existing one ...
      del update_handler
      
      # ... and set the new one.
      update_handler = handler_new
    
  # Clear the events
  events = []
  # Do a display flip.
  pygame.display.flip()
