#
# A python port of the utility TexFont Mark J. Kilgard, 1997.
#
# I'm issuing this under the same licence as the original
#
# This program is freely distributable without licensing fees  and is
# provided without guarantee or warrantee expressed or  implied. This
# program is -not- in the public domain.
#

# Import system libraries
import sys, struct

# PyOpenGL imports
try:
    from OpenGL.GL import *
    from OpenGL.GLU import *
except:
    print 'This entry requires PyOpenGL'
    raise SystemExit

# Flag if we want to used the LuminanceAlpha format
use_LuminanceAlpha = False
use_DisplayLists   = False
use_glGenTextures  = glGenTextures
use_glBindTextures = glBindTexture
#use_GL_INTENSITY4  = GL_INTENSITY4
use_GL_INTENSITY4  = GL_INTENSITY

class Font_GlyphInfo :
  
  # Create a glyph info structure by reading it from a file
  def __init__(self, file, swap_char) :
    
    # Read the basic metrics. Create the unpack format ...
    unpack_format = swap_char + 'HBB'

    # ... compute the size of the structure ...
    unpack_size = struct.calcsize(unpack_format)

    # ... and try to read it from the file.
    unpack_str = file.read(unpack_size)
    
    # If there was a premature end of file ...
    if len(unpack_str) != unpack_size :
       
      # ... throw an exception.
      raise IOError('Premature end of file')
    
    # Unpack the data
    (self.char, self.width, self.height) = \
      struct.unpack(unpack_format, unpack_str)
      
    # Apparently unpack can't handle many values at once. Create the unpack
    #format ...
    unpack_format = swap_char + 'bbbxhh'

    # ... compute the size of the structure ...
    unpack_size = struct.calcsize(unpack_format)

    # ... and try to read it from the file.
    unpack_str = file.read(unpack_size)
    
    # If there was a premature end of file ...
    if len(unpack_str) != unpack_size :
       
      # ... throw an exception.
      raise IOError('Premature end of file')
    
    # Unpack the data
    (self.xoffset, self.yoffset, self.advance, self.x, self.y) = \
      struct.unpack(unpack_format, unpack_str)
    
  # Build the lists of points for the texture and verticies
  def build_vertex(self, tex_width, tex_height) :
    
    # Ensure that the texture size is floating point
    tex_width  = float(tex_width)
    tex_height = float(tex_height)
    
    # Build the verticies
    #self.verticies = [\
    #  [\
    #    [(self.x + 0.5) / tex_width, (self.y + 0.5) / tex_height,],
    #    [float(self.xoffset), float(self.yoffset)]
    #  ], [\
    #    [(self.x + self.width + 0.5) / tex_width, (self.y + 0.5) / tex_height,],
    #    [float(self.xoffset + self.width), float(self.yoffset)]
    #  ], [\
    #    [(self.x + self.width + 0.5) / tex_width, (self.y + self.height + 0.5) / tex_height,],
    #    [float(self.xoffset + self.width), float(self.yoffset + self.height)]
    #  ], [\
    #    [(self.x + 0.5) / tex_width, (self.y + self.height + 0.5) / tex_height,],
    #    [float(self.xoffset), float(self.yoffset + self.height)]
    #  ]\
    #]
    self.verticies = [\
      [\
        [(self.x) / tex_width, (self.y) / tex_height,],
        [float(self.xoffset), float(self.yoffset)]
      ], [\
        [(self.x + self.width) / tex_width, (self.y) / tex_height,],
        [float(self.xoffset + self.width), float(self.yoffset)]
      ], [\
        [(self.x + self.width) / tex_width, (self.y + self.height) / tex_height,],
        [float(self.xoffset + self.width), float(self.yoffset + self.height)]
      ], [\
        [(self.x) / tex_width, (self.y + self.height) / tex_height,],
        [float(self.xoffset), float(self.yoffset + self.height)]
      ]\
    ]
    
  # Render a glyph
  def render(self, offset) :
    glPushMatrix();
    glTranslatef(offset, 0.0, 0.0)
    glBegin(GL_QUADS)
    for vertex in self.verticies :
      glTexCoord2fv(vertex[0])
      glVertex2fv(vertex[1])
    glEnd()
    glPopMatrix();
    return offset + self.advance


class Font :

  # Create a font by loading it.
  def __init__(self, filename) :
    
    # Try to open the file.
    file = open(filename, 'rb')
    
    # Try to read the txf header.
    header = file.read(4)
    
    # If there was a premature end of file or the header was wrong ...
    if len(header) != 4 or header.upper() != '\377TXF' :
      
      # ... throw an exception.
      raise IOError('Not a texture font file')
    
    # Try to read the endianness marker.
    endianness_str = file.read(4)
    
    # If there was a premature end of file ...
    if len(endianness_str) != 4 :
      
      # ... throw an exception.
      raise IOError('Not a texture font file')
   
    # Convert the endianness to an integer
    (endianness,) = struct.unpack('i', endianness_str)
    
    # If the endianness is not valid ...
    if endianness != 0x12345678 and endianness != 0x78563412 :
       
      # ... throw an exception.
      raise IOError('Not a texture font file')

    # If we need to do byte swaping ...
    if endianness == 0x78563412 :
      
      # ... decide which unswap character we need.
      if struct.unpack('<i', endianness_str) == (0x12345678,) :
        swap_char = '<'
      else :
        swap_char = '>'
    else :
      
      # We don't need to swap.
      swap_char = ''
    
    # Read the basic metrics. Create the unpack format ...
    unpack_format = swap_char + '6i'
    
    # ... compute the size of the structure ...
    unpack_size = struct.calcsize(unpack_format)
    
    # ... and try to read it from the file.
    unpack_str = file.read(unpack_size)
    
    # If there was a premature end of file ...
    if len(unpack_str) != unpack_size :
       
      # ... throw an exception.
      raise IOError('Premature end of file')
    
    # Unpack the data
    (format, tex_width, tex_height, max_ascent, max_descent, num_glyphs) = \
      struct.unpack(unpack_format, unpack_str)
      
    #print 'format = ' + str(format)
    #print 'tex_width = ' + str(tex_width)
    #print 'tex_height = ' + str(tex_height)
    #print 'max_ascent = ' + str(max_ascent)
    #print 'max_descent = ' + str(max_descent)
    #print 'num_glyphs = ' + str(num_glyphs)
 
    # Iterate over the number of glyphs ...
    idx = 0
    self.glyphs = {}
    while idx < num_glyphs :
      # ... read each on in ... 
      glyph_info = Font_GlyphInfo(file, swap_char)
      
      # ... build the vertex points ...
      glyph_info.build_vertex(tex_width, tex_height)
      
      # ... and add it to the glyph map.
      self.glyphs[chr(glyph_info.char)] = glyph_info
      #print 'Glyph = ' + str(lyph_info.char)
      # Move onto the next glyph
      idx += 1
      
    # Read the texture data. If we are reading byte format ...
    if format == 0 :
      
      # ... try to read in the texture.
      texture_size = tex_width * tex_height
      texture = file.read(texture_size)
    
      # If there was a premature end of file ...
      if len(texture) != texture_size :
       
        # ... throw an exception.
        raise IOError('Premature end of file')
      
      # If we want to use LuminanceAlpha format ...
      if use_LuminanceAlpha :
        
        # ... unpack it by duplicating. This _must_ be slow!
        texture_new = ''
        for char in texture : texture_new += char * 2
        del texture
        
        # Store the image.
        self.texture = texture_new
        self.texture_size = tex_width * tex_height * 2
        self.texture_format = GL_LUMINANCE_ALPHA
      else :
        
        # Otherwise, just store the image.
        self.texture = texture
        self.texture_size = tex_width * tex_height
        self.texture_format = GL_LUMINANCE
      
      # If we are using bitmap format ...
    elif format == 1 :
      
      # ... compute the stride
      stride = (tex_width + 7) >> 3
      
       # ... try to read in the texture.
      texture_size = stride * tex_height
      texture = file.read(texture_size)
      print 'texture_size = ' + str(texture_size)
      print 'read len = ' + str(len(texture))
      # If there was a premature end of file ...
      if len(texture) != texture_size :
       
        # ... throw an exception.
        raise IOError('Premature end of file')
     
      # If we want to use LuminanceAlpha format ...
      if use_LuminanceAlpha :
        
        # ... uppack into 2 bytes. This _must_ be slow!
        texture_new = ''
        i = 0
        while i < tex_height :
          line = texture[i*stride:]
          j = 0
          mask = 1
          value = ord(line[0])
          while j < tex_width :
            if value & mask :
              texture_new += '\377\377'
            else :
              texture_new += '\000\000'
            mask <<= 1
            if mask > 255 :
              mask = 1
              value = ord(line[(j >> 3)+1])
            j += 1
          i += 1
        
        # Idiot check
        assert (len(texture_new) == tex_width * tex_height * 2), "Unpack failed"
        del texture
        
        # Store the texture.
        self.texture = texture_new
        self.texture_size = tex_width * tex_height * 2
        self.texture_format = GL_LUMINANCE_ALPHA
        
      else :
        
        # ... uppack into 1 byte. This _must_ be slow!
        texture_new = ''
        i = 0
        while i < tex_height :
          line = texture[i*stride:]
          j = 0
          mask = 1
          value = ord(line[0])
          while j < tex_width :
            if mask > 255 :
              mask = 1
              value = ord(line[j >> 3])
            if value & mask :
              texture_new += '\377'
            else :
              texture_new += '\000'
            mask <<= 1
            j += 1
          i += 1
        
        # Idiot check
        assert (len(texture_new) == tex_width * tex_height), "Unpack failed"
        del texture
        
        # Store the texture.
        self.texture = texture_new
        self.texture_size = tex_width * tex_height
        self.texture_format = GL_LUMINANCE
             
    else :
       
      # Bad format, throw an exception.
      raise IOError('Unknown texture format')
    
    # Clear the texture object
    self.texobj = 0
    
    # Store the texture size
    self.tex_width = tex_width
    self.tex_height = tex_height
    self.max_ascent = max_ascent
    self.max_descent = max_descent
  
    
  # Establish the texture.  
  def establish_texture(self, mipmaps) :
    
    # If the font hasn't been established before ...
    if self.texobj == 0 :
      
      # ... create an object to put it in, either a display list ...
      if(use_DisplayLists) :
        self.texobj = glGenLists(1)
        self.use_displaylists = True
      else :
        # ... or a texture object.
        self.texobj = glGenTextures(1)
        self.use_displaylists = False
        
    # Bind the object
    if self.use_displaylists :
      glNewList(self.texobj, GL_COMPILE)
    else :
      glBindTexture(GL_TEXTURE_2D, self.texobj)
      
    # Work out thu internal format.
    if self.texture_format == GL_LUMINANCE_ALPHA :
      format = GL_LUMINANCE_ALPHA
    else :
      format = use_GL_INTENSITY4
    
    # Bulid the texture
    if mipmaps :
      gluBuild2DMipmaps(GL_TEXTURE_2D, format, self.tex_width, \
        self.tex_height, self.texture_format,  GL_UNSIGNED_BYTE, \
        self.texture)
    else :
       glTexImage2D(GL_TEXTURE_2D, format, self.tex_width, \
        self.tex_height, 0, self.texture_format,  GL_UNSIGNED_BYTE, \
        self.texture)
        
    # If we are using display list, stop compiling
    if self.use_displaylists :
      glEndList()
      glCallList(self.texobj)
     
  # Bind the texture 
  def bind_texture(self) :
    
    # If the texture has not been established ...
    if self.texobj == 0 :
      
      # ... establish it (which binds it as well)
      self.establish_texture(True)
    else :
      
      # Otherwise, bind the object
      if self.use_displaylists :
        glCallList(self.texobj)
      else :
        glBindTexture(GL_TEXTURE_2D, self.texobj)
  
  # Render a string.
  def render(self, str) :
    offset = 0
    for char in str :
      offset = self.glyphs[char].render(offset)
  
  # Compute the size of a string
  def get_width(self, str) :
    width = 0
    for char in str :
      width += self.glyphs[char].advance
    return width
  
# Test code
if (__name__ == '__main__') :
   
  # Create a font
  font = Font(sys.argv[1])
   