# PIG Pilot (Texture manager)
# Copyright (C) 2005  Robin J Patenall
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#


# Basic system imports
import sys

# PyOpenGL imports
try:
    from OpenGL.GL import *
    from OpenGL.GLU import *
except:
    print 'This entry requires PyOpenGL'
    raise SystemExit

# Import PIL
import Image

# The dictionary of textures.
textures = {}

# The currently bound texture
texture_bound = None

# The size of the current texture
texture_size = [1, 1]

# Set the format wanted for the texture
def set_format(filename, format) :
  
  # Allow me to access the texture info
  global textures
  
  # If there isn't an entry in the dictionary for the texture ...
  if textures.get(filename) == None : 
    
    # ... set an dictionary with the format specified.
    textures[filename] = {'format' : format}
  else :
    
    # Otherwise just set it in the existing dictionaryp
    textures[filename]['format'] = format


# Ensure that a texture is bound as the current texture.
def bind(filename) :
  
  # Allow me to access the texture info
  global textures, texture_bound, texture_size
  
  # If there isn't an entry in the dictionary for the texture ...
  if textures.get(filename) == None : 
    
    # ... set an empty dictionary.
    texture = textures[filename]={}
  else :
    
    # Get the texture reference.
    texture = textures[filename]
  
  # If there isn't a texture name ...
  if texture.get('name') == None :
    
    # ... we have to load the texture. First load the actual image ...
    image = Image.open(filename)
    
    # If we don't already have a format ...
    if texture.get('format') == None :
      
      # ... set a default.
      texture['format'] = GL_RGBA
    
    # Create the raw data to send to OpenGL.
    image_data = image.tostring("raw", "RGBA", 0, -1)
    
    # Create a texture name
    name = glGenTextures(1)
    
    # Bind the texture and store it.
    glBindTexture(GL_TEXTURE_2D, name)
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
    #glTexImage2D(GL_TEXTURE_2D, 0, texture['format'], image.size[0],
    #             image.size[1], 0, GL_RGBA, GL_UNSIGNED_BYTE, image_data)
    gluBuild2DMipmaps(GL_TEXTURE_2D, texture['format'], image.size[0],
                 image.size[1], GL_RGBA, GL_UNSIGNED_BYTE, image_data)
    
    # Destroy the image data now rather than let it hang around.
    del image_data
    
    # Store the texture info.
    texture['name'] = name
    texture['filename'] = filename
    texture['size'] = image.size
    
    # Store that we have this texture bound
    texture_bound = name
    texture_size = image.size
    
  else :
    
    # We've already loaded the texture. If it isn't currently bound ...
    if texture_bound != texture['name'] :
      
      # ... bind it.
      glBindTexture(GL_TEXTURE_2D, texture['name'])
      texture_bound = texture['name']
      texture_size = texture['size']

# Clear the current binding.
def clear_bind() :
  # Allow me to access the texture info
  global textures, texture_bound, texture_size
  texture_bound = None
  
# Create a texture of a given name and size but don't load it
# Used for the text module.
def create_texture(texture_name, size) :
  
  # Allow me to access the texture info
  global textures, texture_bound, texture_size
  
  # If there isn't an entry in the dictionary for the texture ...
  if textures.get(texture_name) == None : 
    
    # ... set an empty dictionary.
    texture = textures[texture_name]={}
  else :
    
    # Get the texture reference.
    texture = textures[texture_name]
  
  # If there isn't a texture name ...
  if texture.get('name') == None :
      
    # If we don't already have a format ...
    if texture.get('format') == None :
      
      # ... set a default.
      texture['format'] = GL_RGBA
    
    # Create an empty image
    image_data = Numeric.zeros(size, Numeric.Int8).tostring();
    
    # Create a texture name
    name = glGenTextures(1)
  
    # Bind the texture and store it.
    glBindTexture(GL_TEXTURE_2D, name)
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
    glTexImage2D(GL_TEXTURE_2D, 0, texture['format'], size[0], size[1], 0,
                 GL_ALPHA, GL_UNSIGNED_BYTE, image_data)
    
    # Destroy the image data now rather than let it hang around.
    del image_data
    
    # Store the texture info.
    texture['name'] = name
    texture['filename'] = texture_name
    texture['size'] = size
    
    # Store that we have this texture bound
    texture_bound = name
    texture_size = size
  else :
    
    # We've already loaded the texture. If it isn't currently bound ...
    if texture_bound != texture['name'] :
      
      # ... bind it.
      glBindTexture(GL_TEXTURE_2D, texture['name'])
      texture_bound = texture['name']
      texture_size = texture['size']
  

# Return the size of the texure.
def get_size(filename) :
  
  # Allow me to access the texture info
  global textures, texture_bound, texture_size
  
  # Idiot check that the texture has been loaded
  assert (textures.get(filename) != None), "Texture '"+filename+"' not loaded"
  assert (textures[filename].get('size') != None), \
    "Texture '"+filename+"' not loaded"
    
  # Return the texture size
  return textures[filename]['size']


# Return the size of the current texture.
def get_current_size() :
  
  # Allow me to access the texture info
  global texture_size
  
  # Return the size
  return texture_size
